//
// FloatListC.h
//

#ifndef _FloatListC_h_
#define _FloatListC_h_

#include <stream.h>

/*
 * OListC.clas -- class definition for class OListC
 *
 *   This type of list store the members themselves
 *   PListC stores only pointers to the members
 */

#include <stream.h>

class FloatListC {

protected:

   float		*_list;
   unsigned	_count;
   unsigned	_space;
   char		allow_duplicates;
   char		sorted;
   char		adjustGrowth;
   char		autoShrink;
   void		grow(unsigned size=0);
   void		shrink();

public:

   unsigned		GROWTH_AMOUNT;
   const int		NULL_INDEX;
   const unsigned	MAX_GROWTH;

   FloatListC();
   FloatListC(unsigned);
   FloatListC(const FloatListC&);
   virtual ~FloatListC();

// Append another list to this one

   FloatListC&	operator+=(const FloatListC&);

// Copy another list to this one

   FloatListC&	operator= (const FloatListC&);

// Return the position of the specified object.	 Object must be in list

   int		indexOf(const float&) const;

// Return the sorted position of the specified object.	Object does not
//    have to be in the list

   unsigned	orderOf(const float&) const;

//
// Return non-zero if the specified object is in the list
//
   int		includes(const float&) const;

//
// Add the specified object to the list in the requested position
//    The position is ignored if this list is sorted
//
   float		*insert (const float&, unsigned);
   float		*append (const float&);
   float		*prepend(const float&);
   float		*add    (const float&);

//
// Remove the specified object or index from the list
//
   void		remove(unsigned);
   void		remove(const float&);
   void		removeAll();
   void		removeLast();

//
// Return a pointer to the previous object in the list
//
   float		*before(const float&) const;

//
// Functions to support sorting
//
   static int	compare(const void*, const void*);
   void		sort();
   void		sort(int (*)(const void*, const void*));
   void		SetSorted(char);

   inline unsigned	size()		const { return _count; }
   inline unsigned	capacity()	const { return _space; }
   inline float		*start()	const { return _list; }

   void		AllowDuplicates(char);
   void		AutoShrink(char);
   void		SetCapacity(unsigned);

   float		*operator[](unsigned) const;

   float		*First() const;
   float		*Last() const;

   ostream&	printOn(ostream& strm=cout) const;
};

//
// PRINTING
//

inline ostream& operator<<(ostream& strm, const FloatListC& c)
{
   return c.printOn(strm);
}

#endif // _FloatListC_h_
