/*
 *  $Id: ImapServerC.h,v 1.11 2000/08/29 13:05:42 evgeny Exp $
 *  
 *  Copyright (c) 1994 HAL Computer Systems International, Ltd.
 * 
 *          HAL COMPUTER SYSTEMS INTERNATIONAL, LTD.
 *                  1315 Dell Avenue
 *                  Campbell, CA  95008
 *
 * Author: Greg Hilton
 * Contributors: Tom Lang, Frank Bieser, and others
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * http://www.gnu.org/copyleft/gpl.html
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
#ifndef _ImapServerC_h_
#define _ImapServerC_h_

#include <hgl/CharC.h>
#include <hgl/StringC.h>

// Known IMAP protocols
#define     IMAP_PROTOCOL_IMAP2BIS      (1L << 0)
#define     IMAP_PROTOCOL_IMAP4         (1L << 1)
#define     IMAP_PROTOCOL_IMAP4REV1     (1L << 2)

// Known authentication methods
#define     IMAP_AUTH_CLEAR_TEXT        (1L << 0) /* IMAP2 LOGIN */
#define     IMAP_AUTH_ANONYMOUS         (1L << 1) /* anonymous */
#define     IMAP_AUTH_LOGIN             (1L << 2) /* base64-encoded clear text */
#define     IMAP_AUTH_CRAM_MD5          (1L << 3) /* reasonably secure */
#define     IMAP_AUTH_GSSAPI            (1L << 4)
#define     IMAP_AUTH_KERBEROS_V4       (1L << 5)

// Known IMAP server bugs
#define     IMAP_BUG_DUPLICATE_SELECT   (1L << 0)
// more to come ;-)


class CharC;
class StringListC;
class IntListC;
class ImapFolderC;


typedef struct _AuthData
{
   CharC user;
   CharC pass;
} AuthData;

// Return codes of RunCommand2()
typedef enum {
   ImapCommandOK,		// OK :-)
   ImapCommandNO,		// Could not be performed by the server
   ImapCommandBAD,		// Unrecognized by the server
   ImapCommandBYE,		// Server closed connection
   ImapCommandFailed,		// Protocol error and/or our internal bug
   ImapCommandCancelled,	// Cancelled by the user
   ImapCommandNoConnect		// Connection impossible (no network etc)
} ImapCommandReturn;

// Server response types as passed to RunCommand2's hook
typedef enum {
   ImapResponseOK,
   ImapResponseNO,
   ImapResponseBAD,
   ImapResponseCont
} ImapResponse;

// Hook function to be supplied for RunCommand2
typedef Boolean (*RunCommandHook)(
   ImapResponse type,
   CharC response,
   StringListC info,
   StringListC data,
   StringC *reply,
   void *data
);

class ImapServerC {

   Boolean	RetryLogin(char*);
   Boolean	Capability();
   void		CloseSocket();
   Boolean      DoLogin();

public:

   StringC	name;
   StringC	user;
   StringC	pass;
   long		protocols;
   long		authMethods;
   long		bugs;
   int		sock;
   Boolean	connected;
   Boolean	authenticated;
   Boolean	dataPending;
   StringC	folder;
   struct hostent	*host;

   ImapServerC(const char*);
   ~ImapServerC();

   Boolean	Append(CharC, CharC, StringListC&);
   Boolean	Append(CharC, StringListC&, char*, StringListC&);
   ImapCommandReturn	Authenticate(CharC, CharC);
   Boolean      Close(StringListC&);
   Boolean      Connect();
   Boolean	Copy(int, CharC, StringListC&);
   Boolean	Create(CharC, StringListC&);
   Boolean	Delete(CharC, StringListC&);
   Boolean	EstablishSession();
   Boolean	Exists(CharC);
   Boolean	Expunge(StringListC&);
   void		ExpandList(StringListC&, StringListC&);
   Boolean	FetchHdrs(int, ImapFolderC*, int*, char**, char**, StringC&);
   void		FetchFlush( StringC& );
   Boolean	Fetch(int, ImapFolderC*, CharC, char**, char**, StringC&);
   void		GenTag(StringC&);
   Boolean	GetLine(StringC&);
   Boolean	ListMailboxes(CharC, StringListC&, StringListC&);
   Boolean	Logout(StringListC&);
   Boolean	Noop(StringListC&);
   Boolean	PutLine(CharC, Boolean terminate=True);
   Boolean      ReConnect();
   Boolean	Rename(CharC, CharC, StringListC&);
   Boolean	RunCommand(CharC, StringListC&);
   ImapCommandReturn	RunCommand2(CharC, RunCommandHook, void *);
   Boolean	Search(CharC, IntListC&, StringListC&);
   Boolean	Select(CharC, StringListC&);
   Boolean	SetFlags(int, u_int);
   void		Unexpected(CharC);
};

#define IMAP_SERVER_PORT	143

extern ImapServerC	*FindImapServer(char*);
extern void		 CloseImapServerConnections();

#endif // _ImapServerC_h_
